==> moe-0.5.tar.bz2: moe-0.5/buffer.h <==
/*  GNU Moe - My Own Editor
    Copyright (C) 2005 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// There are two uses for 'Points'; cursors and pointers.
// 'cursors' are unbounded non-negative coordinates.
// 'pointers' always point to a valid character in 'Buffer' or to eof.
// They are different because characters can be more than a column wide,
// and cursors can go beyond the last character in a buffer line.
//
struct Point
  {
  int line, col;

  Point( int l = -1, int c = 0 ) throw() : line( l ), col( c ) {}

  Point & operator+=( const Point & p ) throw()
    { line += p.line; col += p.col; return *this; }
  Point & operator-=( const Point & p ) throw()
    { line -= p.line; col -= p.col; return *this; }

  Point operator+( const Point & p ) const throw()
    { Point r( *this ); r += p; return r; }
  Point operator-( const Point & p ) const throw()
    { Point r( *this ); r -= p; return r; }

  bool operator==( const Point & p ) const throw()
    { return ( line == p.line && col == p.col ); }
  bool operator!=( const Point & p ) const throw() { return !( *this == p ); }

  bool operator<( const Point & p ) const throw()
    { return ( line < p.line || ( line == p.line && col < p.col ) ); }
  bool operator>( const Point & p ) const throw()
    { return ( line > p.line || ( line == p.line && col > p.col ) ); }

  bool operator<=( const Point & p ) const throw() { return !( *this > p ); }
  bool operator>=( const Point & p ) const throw() { return !( *this < p ); }

  Point & cut() throw()
    { if( line < 0 ) line = 0; if( col < 0 ) col = 0; return *this; }
  };


// All the lines are newline terminated except the last one, whose
// next-to-last character position ( data[last_line()].end() ) is eof().
//
class Basic_buffer
  {
  std::vector< std::string > data;	// the lines

protected:
  void add_line( const char * buf, const int len ) throw();
  bool detect_and_remove_cr() throw();

public:
  Basic_buffer() : data( 1 ) {}
  Basic_buffer( const Basic_buffer & b, const Point & p1, const Point & p2 ) throw();

  bool blank( const int line ) const throw()
    { Point p( line, 0 ); return ( !pisvalid( p ) || bot( p ) == eol( p ) ); }
  int  characters( const int line ) const throw();
  bool empty() const throw() { return ( data.size() == 1 && !data[0].size() ); }
  int  lines() const throw() { return data.size(); }
  int  last_line() const throw() { return lines() - 1; }

  int  operator[]( const Point & p ) const throw();
  int  pgetc( Point & p ) const throw();	// get and move next
  int  pngetc( Point & p ) const throw();	// move next and get
  int  ppgetc( Point & p ) const throw();	// move prev and get
  bool pnext( Point & p ) const throw();
  bool pprev( Point & p ) const throw();
  bool pseek( Point & p, int n ) const throw();
  bool pvalid( Point & p ) const throw();

  // insert block b.[p1,p2) at p and move next
  bool pputb( Point & p, const Basic_buffer & b, const Point & p1, const Point & p2 ) throw();
  // insert char and move next
  bool pputc( Point & p, const unsigned char ch ) throw();
  // change char and move next
  bool pchgc( Point & p, const unsigned char ch ) throw();
  // delete block at [p1,p2)
  bool pdelb( const Point & p1, const Point & p2 ) throw();

  bool pisbow( const Point & p ) const throw();
  bool piseow( const Point & p ) const throw();
  bool pisvalid( const Point & p ) const throw()
    { return ( p == eof() || ( p.col >= 0 && p.col < characters( p.line ) ) ); }

  Point bof() const throw() { return Point( 0, 0 ); }
  Point eof() const throw() { return Point( last_line(), characters( last_line() ) ); }
  Point bol( const Point & p ) const throw();
  Point eol( const Point & p ) const throw();
  Point bot( const Point & p ) const throw();	// begin of text in line
  Point eot( const Point & p ) const throw();	// end of text in line
  Point bop( const Point & p ) const throw();	// begin of paragraph
  Point eop( const Point & p ) const throw();	// end of paragraph

  bool set_to_matching_delimiter( Point & p, bool forward = true,
                                  const bool force = false ) const throw();
  bool find_text( Point & p, const std::string & text, const bool icase = false ) const throw();
  bool rfind_text( Point & p, const std::string & text, const bool icase = false ) const throw();

  Point to_cursor( const Point & pointer ) const throw();
  Point to_pointer( const Point & cursor, Point & pcursor ) const throw();
  int to_string( const Point & p1, const Point & p2, std::string & text ) const throw();
  };


class Buffer : public Basic_buffer
  {
  bool crlf;
  mutable bool _modified, saved;
  mutable std::string _name;

  struct Change_history
    {
    struct Atom
      {
      Point begin;	// Buffer position of delete or begin of insert
      Point end;		// Buffer position of end of insert
      Basic_buffer *bbp;	// Basic_buffer containing deleted data

      // Creates an insert Atom
      Atom( const Point & p1, const Point & p2 ) throw()
        : begin( p1 ), end( p2 ), bbp( 0 ) {}

      // Creates a delete Atom
      Atom( const Point & p1, const Point & p2, const Basic_buffer & b ) throw()
        : begin( p1 ), end( p1 )
        { bbp = new Basic_buffer( b, p1, p2 ); }

      // Creates a change (overwrite) Atom
      Atom( const Point & p1, const Point & p2, const Basic_buffer & b,
            const Point & bp1, const Point & bp2 ) throw()
        : begin( p1 ), end( p2 )
        { bbp = new Basic_buffer( b, bp1, bp2 ); }
      };

    struct Record
      {
      Point pointer;		// Pointer position before the change
      mutable bool modified;	// Status of modified flag before this record
      std::vector< Atom > atom_vector;

      Record( const Point & p, const bool m ) throw()
        : pointer( p ), modified( m ) {}
      void add_atom( const Atom & a ) throw() { atom_vector.push_back( a ); }
      };

    int index;			// Index of record to be undone
    mutable bool appendable;	// True if last record added is appendable
    mutable bool force_append;
    std::vector< Record > record_vector;

    Change_history() throw()
      : index( -1 ), appendable( false ), force_append( false ) {}
    ~Change_history() throw();

    void add_record( const Basic_buffer & buffer, const Record & r ) throw();
    bool append_record( const Basic_buffer & buffer, const Record & r ) throw();
    int records() const throw() { return record_vector.size(); }
    };

  Change_history change_history;

public:
  struct Error
    {
    const char * s;
    Error( const char * p ) { s = p; }
    };

  class Options
    {
    int _lmargin, _rmargin;
  public:
    bool auto_indent, overwrite, read_only, word_wrap;

    Options() throw()
      : _lmargin( 0 ), _rmargin( 71 ), auto_indent( false ),
      overwrite( false ), read_only( false ), word_wrap( false ) {}

    void reset() throw() { *this = Options(); }
    int lmargin() const throw() { return _lmargin; }
    int rmargin() const throw() { return _rmargin; }
    bool set_lmargin( const int l ) throw();
    bool set_rmargin( const int r ) throw();
    };

  Options options;
  Point mark[10];		// user bookmarks

  Buffer( const Options & opts, const std::string * namep = 0 ) throw( Buffer::Error );
  bool save( const std::string * namep = 0, bool set_name = true, bool append = false,
             Point p1 = Point(), Point p2 = Point() ) const throw( Buffer::Error );

  const std::string & character_info( const Point & p ) const throw();
  bool modified() const throw() { return _modified; }
  const std::string & name() const throw() { return _name; }
  const std::string & uname() const throw();

       // insert block b.[p1,p2) at p and move next
  bool pputb( Point & p, const Basic_buffer & b, const Point & p1,
              const Point & p2, const bool save_change = true ) throw();
       // put char and move next
  bool pputc( Point & p, const unsigned char ch ) throw();

       // delete block at [p1,p2)
  bool pdelb( const Point & p1, const Point & p2, const int save_change = 1 ) throw();
       // delete char at 'p', or at '--p' if back is true
  bool pdelc( Point & p, const bool back = false ) throw();
  bool pdell( Point & p ) throw();	// delete line and set 'p' to bol()

  bool undo( Point & p ) throw();
  bool redo( Point & p ) throw();
  void force_append() const throw() { change_history.force_append = true; }
  void reset_appendable() const throw() { change_history.appendable = false; }
  };


class Buffer_handle
  {
  Buffer * _bufferp;	// handled buffer

public:
  Point top_left;	// buffer coordinates of window top-left corner
  Point cursor;		// buffer coordinates of window cursor
  Point pointer;	// current buffer character
  Point pcursor;	// buffer coordinates of current buffer character

  Buffer_handle( Buffer & buffer, const int line = 0 ) throw()
    : _bufferp( &buffer ), top_left( line, 0 ), cursor( line, 0 ),
      pointer( line, 0 ), pcursor( line, 0 ) {}

  Buffer & buffer() throw() { return *_bufferp; }
  const Buffer & buffer() const throw() { return *_bufferp; }
  };


// Bufhandle_vector is a circular buffer of Buffer_handles
//
class Bufhandle_vector
  {
  int _first;				// index of first element
  std::vector< Buffer_handle > data;

public:
  Bufhandle_vector() throw() : _first( 0 ) {}
  ~Bufhandle_vector() throw();
  bool save_all_named() const throw();

  Buffer_handle & operator[]( int i ) throw()
    { return data[( _first + i ) % handles()]; }
  int handles() const throw() { return data.size(); }
  int handles( const Buffer & buffer ) const throw();
  void next() throw() { _first = ++_first % handles(); }
  void prev() throw() { _first = ( --_first + handles() ) % handles(); }

  int add_handle( const Buffer::Options & opts, const std::string * namep = 0,
                  int i = -1, const int line = 0 ) throw( Buffer::Error );
  void add_handle_if_pending_input_or_empty( const int line ) throw( Buffer::Error );
  void duplicate_handle( int i ) throw();
  bool delete_handle( int i ) throw();
  };


// Main data structure of the editor.
// Contains the handles to all the buffers being edited.
//
extern Bufhandle_vector bufhandle_vector;

==> moe-0.6.tar.bz2: moe-0.6/buffer.h <==
/*  GNU Moe - My Own Editor
    Copyright (C) 2005, 2006 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// There are two uses for 'Points'; cursors and pointers.
// 'cursors' are unbounded non-negative coordinates.
// 'pointers' always point to a valid character in 'Buffer' or to eof.
// They are different because characters can be more than a column wide,
// and cursors can go beyond the last character in a buffer line.
//
struct Point
  {
  int line, col;

  Point( int l = -1, int c = 0 ) throw() : line( l ), col( c ) {}

  Point & operator+=( const Point & p ) throw()
    { line += p.line; col += p.col; return *this; }
  Point & operator-=( const Point & p ) throw()
    { line -= p.line; col -= p.col; return *this; }

  Point operator+( const Point & p ) const throw()
    { Point r( *this ); r += p; return r; }
  Point operator-( const Point & p ) const throw()
    { Point r( *this ); r -= p; return r; }

  bool operator==( const Point & p ) const throw()
    { return ( line == p.line && col == p.col ); }
  bool operator!=( const Point & p ) const throw() { return !( *this == p ); }

  bool operator<( const Point & p ) const throw()
    { return ( line < p.line || ( line == p.line && col < p.col ) ); }
  bool operator>( const Point & p ) const throw()
    { return ( line > p.line || ( line == p.line && col > p.col ) ); }

  bool operator<=( const Point & p ) const throw() { return !( *this > p ); }
  bool operator>=( const Point & p ) const throw() { return !( *this < p ); }

  Point & cut() throw()
    { if( line < 0 ) line = 0; if( col < 0 ) col = 0; return *this; }
  };


// All the lines are newline terminated except the last one, whose
// next-to-last character position ( data[last_line()].end() ) is eof().
//
class Basic_buffer
  {
  std::vector< std::string > data;	// the lines

protected:
  void add_line( const char * buf, const int len ) throw();
  bool detect_and_remove_cr() throw();

public:
  Basic_buffer() : data( 1 ) {}
  Basic_buffer( const Basic_buffer & b, const Point & p1, const Point & p2 ) throw();

  bool blank( const int line ) const throw()
    { Point p( line, 0 ); return ( !pisvalid( p ) || bot( p ) == eol( p ) ); }
  int  characters( const int line ) const throw();
  bool empty() const throw() { return ( data.size() == 1 && !data[0].size() ); }
  int  lines() const throw() { return data.size(); }
  int  last_line() const throw() { return lines() - 1; }
  bool one_character() const throw();		// buffer has only one character

  int  operator[]( const Point & p ) const throw();
  int  pgetc( Point & p ) const throw();	// get and move next
  int  pngetc( Point & p ) const throw();	// move next and get
  int  ppgetc( Point & p ) const throw();	// move prev and get
  bool pnext( Point & p ) const throw();
  bool pprev( Point & p ) const throw();
  bool pseek( Point & p, int n ) const throw();
  bool pvalid( Point & p ) const throw();

  // insert block b.[p1,p2) at p and move next
  bool pputb( Point & p, const Basic_buffer & b, const Point & p1, const Point & p2 ) throw();
  // insert char and move next
  bool pputc( Point & p, const unsigned char ch ) throw();
  // change char and move next
  bool pchgc( Point & p, const unsigned char ch ) throw();
  // delete block at [p1,p2)
  bool pdelb( const Point & p1, const Point & p2 ) throw();

  bool pisbow( const Point & p ) const throw();
  bool piseow( const Point & p ) const throw();
  bool pisvalid( const Point & p ) const throw()
    { return ( ( p.col >= 0 && p.col < characters( p.line ) ) || p == eof() ); }

  Point bof() const throw() { return Point( 0, 0 ); }
  Point eof() const throw() { return Point( last_line(), data[last_line()].size() ); }
  Point bol( const Point & p ) const throw();
  Point eol( const Point & p ) const throw();
  Point bot( const Point & p ) const throw();	// begin of text in line
  Point eot( const Point & p ) const throw();	// end of text in line
  Point bop( const Point & p ) const throw();	// begin of paragraph
  Point eop( const Point & p ) const throw();	// end of paragraph

  bool set_to_matching_delimiter( Point & p, bool forward = true,
                                  const bool force = false ) const throw();
  bool find_text( Point & p, const std::string & text, const bool icase = false ) const throw();
  bool rfind_text( Point & p, const std::string & text, const bool icase = false ) const throw();

  Point to_cursor( const Point & pointer ) const throw();
  Point to_pointer( const Point & cursor, Point & pcursor ) const throw();
  int to_string( const Point & p1, const Point & p2, std::string & text ) const throw();
  };


class Buffer : public Basic_buffer
  {
  bool crlf;
  mutable bool _modified, saved;
  mutable std::string _name;

  struct Change_history
    {
    struct Atom
      {
      Point begin;	// Buffer position of delete or begin of insert
      Point end;		// Buffer position of end of insert
      Basic_buffer *bbp;	// Basic_buffer containing deleted data

      // Creates an insert Atom
      Atom( const Point & p1, const Point & p2 ) throw()
        : begin( p1 ), end( p2 ), bbp( 0 ) {}

      // Creates a delete Atom
      Atom( const Point & p1, const Point & p2, const Basic_buffer & b ) throw()
        : begin( p1 ), end( p1 )
        { bbp = new Basic_buffer( b, p1, p2 ); }

      // Creates a change (overwrite) Atom
      Atom( const Point & p1, const Point & p2, const Basic_buffer & b,
            const Point & bp1, const Point & bp2 ) throw()
        : begin( p1 ), end( p2 )
        { bbp = new Basic_buffer( b, bp1, bp2 ); }
      };

    struct Record
      {
      Point before, after;	// Pointer position before and after the change
      mutable bool modified;	// Status of modified flag before this record
      std::vector< Atom > atom_vector;

      Record( const Point & pb, const Point & pa, const bool m ) throw()
        : before( pb ), after( pa ), modified( m ) {}
      void add_atom( const Atom & a ) throw() { atom_vector.push_back( a ); }
      bool append_record( const Basic_buffer & buffer, Record & r ) throw();
      void delete_atoms() throw();
      };

    int index;			// Index of record to be undone
    mutable bool appendable;	// True if last record added is appendable
    mutable bool force_append;
    std::vector< Record > record_vector;

    Change_history() throw()
      : index( -1 ), appendable( false ), force_append( false ) {}
    ~Change_history() throw();

    void add_record( const Basic_buffer & buffer, Record & r ) throw();
    int records() const throw() { return record_vector.size(); }
    };

  Change_history change_history;

public:
  struct Error
    {
    const char * s;
    Error( const char * p ) { s = p; }
    };

  class Options
    {
    int _lmargin, _rmargin;
  public:
    bool auto_indent, overwrite, read_only, word_wrap;

    Options() throw()
      : _lmargin( 0 ), _rmargin( 71 ), auto_indent( false ),
      overwrite( false ), read_only( false ), word_wrap( false ) {}

    void reset() throw() { *this = Options(); }
    int lmargin() const throw() { return _lmargin; }
    int rmargin() const throw() { return _rmargin; }
    bool set_lmargin( const int l ) throw();
    bool set_rmargin( const int r ) throw();
    };

  Options options;
  Point mark[10];		// user bookmarks

  Buffer( const Options & opts, const std::string * namep = 0 ) throw( Buffer::Error );
  bool save( const std::string * namep = 0, bool set_name = true, bool append = false,
             Point p1 = Point(), Point p2 = Point() ) const throw( Buffer::Error );

  const std::string & character_info( const Point & p ) const throw();
  bool modified() const throw() { return _modified; }
  const std::string & name() const throw() { return _name; }
  const std::string & uname() const throw();

       // insert block b.[p1,p2) at p and move next
  bool pputb( Point & p, const Basic_buffer & b, const Point & p1,
              const Point & p2, const int save_change = 1 ) throw();
       // put char and move next
  bool pputc( Point & p, const unsigned char ch ) throw();

       // delete block at [p1,p2)
  bool pdelb( const Point & p1, const Point & p2, const int save_change = 1 ) throw();
       // delete char at 'p', or at '--p' if back is true
  bool pdelc( Point & p, const bool back = false ) throw();
  bool pdell( Point & p ) throw();	// delete line and set 'p' to bol()

  bool undo( Point & p ) throw();
  bool redo( Point & p ) throw();
  void force_append() const throw() { change_history.force_append = true; }
  void reset_appendable() const throw() { change_history.appendable = false; }
  };


class Buffer_handle
  {
  Buffer * _bufferp;	// handled buffer

public:
  Point top_left;	// buffer coordinates of window top-left corner
  Point cursor;		// buffer coordinates of window cursor
  Point pointer;	// current buffer character
  Point pcursor;	// buffer coordinates of current buffer character

  Buffer_handle( Buffer & buffer, const int line = 0 ) throw()
    : _bufferp( &buffer ), top_left( line, 0 ), cursor( line, 0 ),
      pointer( line, 0 ), pcursor( line, 0 ) {}

  Buffer & buffer() throw() { return *_bufferp; }
  const Buffer & buffer() const throw() { return *_bufferp; }
  };


// Bufhandle_vector is a circular buffer of Buffer_handles
//
class Bufhandle_vector
  {
  int _first;				// index of first element
  std::vector< Buffer_handle > data;

public:
  Bufhandle_vector() throw() : _first( 0 ) {}
  ~Bufhandle_vector() throw();
  bool save_all_named() const throw();

  Buffer_handle & operator[]( int i ) throw()
    { return data[( _first + i ) % handles()]; }
  int handles() const throw() { return data.size(); }
  int handles( const Buffer & buffer ) const throw();
  void next() throw() { _first = ++_first % handles(); }
  void prev() throw() { _first = ( --_first + handles() ) % handles(); }

  int find_or_add_handle( const Buffer::Options & opts,
                          const std::string * namep = 0,
                          int i = -1, const int line = 0 ) throw( Buffer::Error );
  void add_handle_if_pending_input_or_empty( const int line ) throw( Buffer::Error );
  void duplicate_handle( int i ) throw();
  bool delete_handle( int i ) throw();
  };


// Main data structure of the editor.
// Contains the handles to all the buffers being edited.
//
extern Bufhandle_vector bufhandle_vector;
